/*
   EventDB.c : Event database engine for trace toolkit.
   Copyright (C) 1999 Karim Yaghmour.
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   History : 
     K.Y., 11/10/99, Modified to conform with complete kernel instrumentation
     K.Y., 26/06/99, Initial typing.
*/

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "Tables.h"
#include "FileIO.h"
#include "EventDB.h"

/******************************************************************
 * Function :
 *    DBReadTrace()
 * Description :
 *    Fills the trace data-base using the raw data file and
 *    according to the options provided at the command line.
 * Parameters :
 *    pmTraceDataFile, File containing the raw trace data.
 *    pmTraceDB, The event database to be built.
 * Return values :
 *    TRUE, The event database was correctly read.
 *    FALSE, The input file doesn't contain a valid trace.
 * History :
 *    K.Y., 20/05/99, Initial typing.
 ******************************************************************/
int DBReadTrace(int pmTraceDataFile, db* pmTraceDB)
{
  int               lFirstEvent = TRUE;     /* Is it the first event read */
  int               lEventID;               /* Identy of event */
  int               lEventDataSize;         /* Size of event data */
  char*             lEventString;           /* String describing the event */
  event*            pEvent;                 /* Pointer on event */
  event*            pEventLast;             /* Last event added */
  struct timeval    lEventTime;             /* The time at which the event occured */

  /* Initialize database */
  pmTraceDB->Nb               = 0;
  pmTraceDB->Properties.State = UNINIT;
  pmTraceDB->Events           = NULL;
  memset(&(pmTraceDB->StartTime), 0, sizeof(struct timeval));
  memset(&(pmTraceDB->EndTime), 0, sizeof(struct timeval));

  /* Set up the first element */
  pmTraceDB->Events = (event*) malloc(sizeof(event));
  pEvent = pmTraceDB->Events;
  pEvent->Prev = NULL;

  /* Set last element */
  pEventLast = NULL;

  /* Read it through */
  while(ReadRawEntry(pmTraceDataFile,
		     &(pmTraceDB->Properties),
		     &(pEvent->ID),
		     &(pEvent->Time),
		     &(pEvent->CPUID),
		     &(pEvent->Struct),
		     &(pEvent->String),
		     &(pEvent->Size)) == TRUE)
    {
    /* Has the first event initialized the structures properly */
    if((lFirstEvent == TRUE) && (pmTraceDB->Properties.State != INIT))
      {
      /* Free the first event */
      free(pmTraceDB->Events);

      /* Reset first event */
      pmTraceDB->Events = NULL;
    
      /* Tell the caller that something is wrong */
      return FALSE;     
      }
    else
      lFirstEvent = FALSE;

    /* Update the number of events */
    pmTraceDB->Nb++;

    /* Set event's owner */
    pEvent->Process = NULL;

    /* Set last event */
    pEventLast = pEvent;

    /* Allocate a new event, link it and go to it */
    pEvent->Next = (event*) malloc(sizeof(event));
    pEvent->Next->Prev = pEvent;
    pEvent = pEvent->Next;

    /* Copy last time to new event structure so that it would be updated */
    memcpy(&(pEvent->Time),
	   &(pEventLast->Time),
	   sizeof(pEvent->Time));
    }

  /* Did we get to read anything */
  if(pEventLast != NULL)
    {
    /* Last malloc is always wasted since we stop when ReadRawEntry returns FALSE */
    pEventLast->Next = NULL;
    free(pEvent);

    /* Set last event */
    pmTraceDB->LastEvent = pEventLast;

    /* Set the start and end time */
    memcpy(&(pmTraceDB->StartTime), &(pmTraceDB->Events->Time), sizeof(struct timeval));
    memcpy(&(pmTraceDB->EndTime), &(pEventLast->Time), sizeof(struct timeval));
    }
  else
    {
    /* Free the first event */
    free(pmTraceDB->Events);

    /* Reset first event */
    pmTraceDB->Events = NULL;

    /* Tell the caller that something is wrong */
    return FALSE;
    }

  /* Tell the caller that everything is OK */
  return TRUE;
}

/******************************************************************
 * Function :
 *    DBProcessProcInfo()
 * Description :
 *    Reads the /proc info file and fills the system process list
 *    accordingly.
 * Parameters :
 *    pmProcDataFile, File pointer to /proc info file.
 *    pmSystem, Description of the system to be filled.
 * Return values :
 *    NONE
 * History :
 *    K.Y., 27/05/99, Initial typing.
 * Note :
 ******************************************************************/
void DBProcessProcInfo(FILE* pmProcDataFile, system* pmSystem)
{
  int        lPID, lPPID;            /* Process PID and Parent PID */
  char       lName[256];             /* Process name */
  process*   pProcess;               /* Generic process pointer */

  /* Initialize system process list  */
  pmSystem->Processes    = NULL;

  /* Read it through */
  while(fscanf(pmProcDataFile, "PID: %d; PPID: %d; NAME: %s\n", &lPID, &lPPID, lName) > 0)
    {
    /* Create a new process */
    pProcess = DBCreateProcess(pmSystem, lPID, lPPID);

    /* Set the command line */
    pProcess->Command   = (char*) malloc(strlen(lName) + 1);
    strcpy(pProcess->Command, lName);
    }
}

/******************************************************************
 * Function :
 *    DBFindProcInTree()
 * Description :
 *    Finds a process with given PID in a process tree.
 * Parameters :
 *    pmPID, PID of process to be found.
 *    pmTree, Tree to be searched.
 * Return values :
 *    Valid pointer to process, if process found.
 *    NULL, otherwise
 * History :
 *    K.Y., 17/06/99, Initial typing.
 * Note :
 *    This function is recursive.
 ******************************************************************/
process* DBFindProcInTree(int pmPID, process* pmTree)
{
  process* pProcJ;    /* Generic process pointers */
  process* pProcess;  /* Process found */

  /* For all the children in the same level */
  for(pProcJ = pmTree; pProcJ != NULL; pProcJ = pProcJ->NextChild)
    {
    /* Is this the process we're looking for */
    if(pProcJ->PID == pmPID)
      return pProcJ;

    /* Does this process have children of his own  */
    if(pProcJ->Children != NULL)
      /* Give this part of the tree to a recursive copy of this function */
      if((pProcess = DBFindProcInTree(pmPID, pProcJ->Children)) != NULL)
	return pProcess;
    }

  /* We didn't find anything */
  return NULL;
}

/******************************************************************
 * Function :
 *    DBBuildProcTree()
 * Description :
 *    Builds the process tree for a system description.
 * Parameters :
 *    pmSystem, The system who's proc tree has to be built.
 * Return values :
 *    NONE
 * History :
 *    K.Y., 17/06/99, Initial typing.
 * Note :
 ******************************************************************/
void DBBuildProcTree(system* pmSystem)
{
  process*    pProcI;         /* Generic process pointer */
  process*    pProcJ;         /* Generic process pointer */
  process*    pPrevProc;      /* Previous process in process list */
  process*    pTopParent;     /* The current topmost parent */
  process*    pProcess;       /* Process found in process tree */
  process*    pProcTemp;      /* Temporary process pointer */

  /* Go through process list */
  for(pProcI = pmSystem->Processes; pProcI != NULL; pProcI = pProcI->Next)
    {
    /* Is this the first process */
    if(pmSystem->ProcTree == NULL)
      {
      /* Set the first element in tree */
      pmSystem->ProcTree = pProcI;
      
      /* Go on to the rest of the processes */
      continue;
      }

    /* Is this process parent any of the topmost process */
    pPrevProc = NULL;
    for(pProcJ = pmSystem->ProcTree; pProcJ != NULL;)
      {
      /* Is this a child of the current process */
      if(pProcI->PID == pProcJ->PPID)
	{
	/* Remember next element */
	pProcTemp = pProcJ->NextChild;

	/* Remove this child from the topmost list */
	if(pPrevProc == NULL)
	  pmSystem->ProcTree = pProcJ->NextChild;
	else
	  pPrevProc->NextChild = pProcJ->NextChild;

	/* Insert this in front of this parent's children list */
	pProcJ->NextChild = pProcI->Children;
	pProcI->Children  = pProcJ;

	/* Update process pointer */
	pProcJ = pProcTemp;
	}
      else
	{
	/* Remember current position and update to next position */
	pPrevProc = pProcJ;
	pProcJ = pProcJ->NextChild;
	}
      }

    /* Is the parent of this process already in the tree */
    if((pProcess = DBFindProcInTree(pProcI->PPID, pmSystem->ProcTree)) != NULL)
      {
      /* Place the new child in front of the children list */
      pProcI->NextChild  = pProcess->Children;
      pProcess->Children = pProcI;
      }
    else
      {
      /* Place this process as the first child at the topmost level */
      pProcI->NextChild  = pmSystem->ProcTree;
      pmSystem->ProcTree = pProcI;
      }
    }
}

/******************************************************************
 * Function :
 *    DBProcessTrace()
 * Description :
 *    Go through the database and enact the options specified while
 *    filling in the description of the system during the trace.
 * Parameters :
 *    pmSystem, Description of the system to be filled
 *    pmTraceDB, The event database to be processed.
 *    pmOptions, The user given options.
 * Return values :
 *    NONE
 * History :
 *    K.Y., 20/05/99, Initial typing.
 * Note :
 ******************************************************************/
void DBProcessTrace(system* pmSystem, db* pmTraceDB, options* pmOptions)
{
  int            lOPID;           /* Outgoing process's PID */
  int            lPID = -1;       /* Process ID */
  char           lProcName[100];  /* The name of the process */
  long           lSec, lUsec;     /* Seconds, microseconds used to calculate durations */
  event*         pEvent;          /* Generic pointer to an event */
  event*         pLastCtrlEvent;  /* Last system-wide control event */
  syscall*       pSyscall;        /* System call pointer */
  syscall*       pPrevSyscall;    /* Previous system call in syscall list */
  syscall*       pSyscallP;       /* Processed: System call pointer */
  process*       pProcess = NULL; /* Current process */
  process*       pProcessChild;   /* Child process */
  struct timeval lTime;           /* Local time variable */

  /* Initialize system description */
  pmSystem->SyscallEntry = 0;
  pmSystem->SyscallExit  = 0;
  pmSystem->TrapEntry    = 0;
  pmSystem->TrapExit     = 0;
  pmSystem->IRQEntry     = 0;
  pmSystem->IRQExit      = 0;
  pmSystem->SchedChange  = 0;
  pmSystem->KernelTimer  = 0;
  pmSystem->BH           = 0;
  pmSystem->TimerExpire  = 0;
  pmSystem->PageAlloc    = 0;
  pmSystem->PageFree     = 0;
  pmSystem->PacketOut    = 0;
  pmSystem->PacketIn     = 0;

  /* Initialize system structures */
  pmSystem->Syscalls     = NULL;
  if(pmOptions->ProcFileName == NULL)
    pmSystem->Processes    = NULL;
  pmSystem->ProcTree     = NULL;

  /* Initialize last control event */
  pLastCtrlEvent = NULL;

  /* Go through the trace database and fill system description accordingly */
  for(pEvent = pmTraceDB->Events; pEvent != NULL; pEvent = pEvent->Next)
    {
    /* Forget what the trace daemon does right after opening the trace device */
    if((lPID == -1) && (pEvent->ID != TRACE_SCHEDCHANGE))
      continue;

    /* Set event's owner */
    pEvent->Process = pProcess;

    /* Initialize event as not being a control event */
    pEvent->Control = FALSE;

    /* Are the details of the event present */
    if(!test_bit(pEvent->ID, &(pmTraceDB->Properties.DetailsMask)))
      continue;

    /* Is this an exiting control event */
    if((pEvent->ID == TRACE_SYSCALL_EXIT)
     ||(pEvent->ID == TRACE_TRAP_EXIT)
     ||(pEvent->ID == TRACE_IRQ_EXIT)
#if 0 /* This is done later since pProcess == NULL for the first incoming process */
     ||((pEvent->ID == TRACE_SCHEDCHANGE)
      &&(SCHED_EVENT(pEvent->Struct)->in != 0))
#endif
     ||(pEvent->ID == TRACE_KERNEL_TIMER)
     ||(pEvent->ID == TRACE_BOTTOM_HALF))
      {
      /* Is this a control event */
      if((pEvent->Next != NULL)
       &&(pProcess != NULL)
       &&(pProcess->PID != 0)
       &&((pEvent->Next->ID == TRACE_SYSCALL_ENTRY)
        ||(pEvent->Next->ID == TRACE_TRAP_ENTRY)
        ||(pEvent->Next->ID == TRACE_IRQ_ENTRY)))
	{
	pEvent->Control = TRUE;
	pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
	}
      }

    /* Depending on the type of event */
    switch(pEvent->ID)
      {
      /* Entry in a given system call */
      case TRACE_SYSCALL_ENTRY :
	/* Increment number of system call entries */
	pProcess->NbSyscalls++;
	pmSystem->SyscallEntry++;

	/* Set system call depth */
	pProcess->Depth++;

	/* Create new pending system call */
	if((pSyscall = (syscall*) malloc(sizeof(syscall))) == NULL)
	  {
	  printf("Memory allocation problem \n");
	  exit(1);
	  }

	/* Set new system call */
	pSyscall->ID    = SYSCALL_EVENT(pEvent->Struct)->syscall_id;
	pSyscall->Nb    = 0;
	pSyscall->Depth = pProcess->Depth;
	memcpy(&(pSyscall->Time), &(pEvent->Time), sizeof(struct timeval));

	/* Link this syscall to the others pending */
	pSyscall->Next = pProcess->Pending;
	pProcess->Pending = pSyscall;

#if 0
	/* Is this the first control event */
	if(pProcess->LastCtrlEvent != NULL)
	  {
	  /* Was the last control event other than an exit */
	  if((pProcess->LastCtrlEvent->ID == TRACE_SYSCALL_ENTRY)
	   ||(pProcess->LastCtrlEvent->ID == TRACE_TRAP_ENTRY)
	   ||(pProcess->LastCtrlEvent->ID == TRACE_IRQ_ENTRY))
	    /* We entered something while we were already in something else */
	    pProcess->CtrlDepth++;
	  else
	    {
	    /* Accumulate the time that passed since the last control event */
	    DBTimeSub(lTime, pEvent->Time, pProcess->LastCtrlEvent->Time);
	    DBTimeAdd(pProcess->TimeRuning, pProcess->TimeRuning, lTime);
	    }
	  }

	/* System call entries are always a control event */
	pEvent->Control = TRUE;
	pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
#else
	/* Is this the first control event */
	if(pLastCtrlEvent != NULL)
	  {
	  /* Was the last control event other than an exit */
	  if((pLastCtrlEvent->ID == TRACE_SYSCALL_EXIT)
	   ||(pLastCtrlEvent->ID == TRACE_TRAP_EXIT)
	   ||(pLastCtrlEvent->ID == TRACE_IRQ_EXIT)
	   ||(pLastCtrlEvent->ID == TRACE_SCHEDCHANGE)
	   ||(pLastCtrlEvent->ID == TRACE_KERNEL_TIMER)
	   ||(pLastCtrlEvent->ID == TRACE_BOTTOM_HALF))
	    {
	    /* Accumulate the time that passed since the last control event */
	    DBTimeSub(lTime, pEvent->Time, pLastCtrlEvent->Time);
	    DBTimeAdd(pProcess->TimeRuning, pProcess->TimeRuning, lTime);

	    /* This is a control event */
	    pEvent->Control = TRUE;
	    pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
	    }
	  }
#if 0
	else
	  {
	  /* This is a control event */
	  pEvent->Control = TRUE;
	  pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
	  }
#endif
#endif
	break;
      
      /* Exit from a given system call */
      case TRACE_SYSCALL_EXIT :
	/* Increment number of system call exits */
	pmSystem->SyscallExit++;

	/* Find the system call in the list of pending system calls */
	pPrevSyscall = pProcess->Pending;
	for(pSyscall = pProcess->Pending; pSyscall != NULL; pSyscall = pSyscall->Next)
	  {
	  if(pSyscall->Depth == pProcess->Depth)
	    break;
	  pPrevSyscall = pSyscall;
	  }

	/* Paranoia */
	if(pSyscall == NULL)
	  {
	  /* The first time the trace daemon leaves a system call, the entry wasn't noted
	   because the daemon wasn't started*/
	  continue;
#if 0
	  printf("Internal error: Leaving unknow system call depth \n");
	  exit(1);
#endif
	  }

	/* Remove syscall from pending syscalls */
	pPrevSyscall->Next = pSyscall->Next;

	/* Find syscall in already processed syscalls list */
	for(pSyscallP = pProcess->Syscalls; pSyscallP != NULL; pSyscallP = pSyscallP->Next)
	  if(pSyscallP->ID == pSyscall->ID)
	    break;

	/* Did we find anything */
	if(pSyscallP == NULL)
	  {
	  /* Allocate space for new syscall */
	  if((pSyscallP = (syscall*) malloc(sizeof(syscall))) == NULL)
	    {
	    printf("Memory allocation problem \n");
	    exit(1);
	    }

	  /* Set and link new syscall */
	  pSyscallP->ID    = pSyscall->ID;
	  pSyscallP->Nb    = 0;
	  pSyscallP->Depth = 0;
	  memset(&(pSyscallP->Time), 0, sizeof(struct timeval));
	  pSyscallP->Next = pProcess->Syscalls;
	  pProcess->Syscalls = pSyscallP;
	  }

	/* Increment number of times syscall was called */
	pSyscallP->Nb++;

	/* Add time spent in this syscall */
	lSec = pEvent->Time.tv_sec - pSyscall->Time.tv_sec;
	lUsec = pEvent->Time.tv_usec - pSyscall->Time.tv_usec;
	if(lUsec < 0)
	  {
	  lSec--;
	  lUsec = 1000000 + lUsec;
	  }
	pSyscallP->Time.tv_sec += lSec;
	pSyscallP->Time.tv_usec += lUsec;

	/* Have we more than a million microseconds */
	if(pSyscallP->Time.tv_usec > 1000000)
	  {
	  pSyscallP->Time.tv_sec += pSyscallP->Time.tv_usec / 1000000;
	  pSyscallP->Time.tv_usec = pSyscallP->Time.tv_usec % 1000000;
	  }

	/* Decrement syscall depth */
	pProcess->Depth--;
	if(pSyscall == pProcess->Pending) pProcess->Pending = NULL;
	free(pSyscall);
	pSyscall = NULL;
	break;

      /* Entry into a trap */
      case TRACE_TRAP_ENTRY :
	pProcess->NbTraps++;
	pmSystem->TrapEntry++;

#if 0
	/* Is this a control event */
	if((pEvent->Prev != NULL)
	 &&(pEvent->Prev->Control == TRUE)
	 &&(pProcess->PID != 0)) 
	  {
	  /* Is this the first control event */
	  if(pProcess->LastCtrlEvent != NULL)
	    {
	    /* Was the last control event other than an exit */
	    if((pProcess->LastCtrlEvent->ID == TRACE_SYSCALL_ENTRY)
	     ||(pProcess->LastCtrlEvent->ID == TRACE_TRAP_ENTRY)
	     ||(pProcess->LastCtrlEvent->ID == TRACE_IRQ_ENTRY))
	      /* We entered something while we were already in something else */
	      pProcess->CtrlDepth++;
	    else
	      {
	      /* Accumulate the time that passed since the last control event */
	      DBTimeSub(lTime, pEvent->Time, pProcess->LastCtrlEvent->Time);
	      DBTimeAdd(pProcess->TimeRuning, pProcess->TimeRuning, lTime);
	      }
	    }

	  /* Trap entries are always a control event */
	  pEvent->Control = TRUE;
	  pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
	  }
#else
	/* Is this the first control event */
	if((pLastCtrlEvent != NULL)
	 &&(pProcess->PID != 0)) 
	  {
	  /* Was the last control event other than an exit */
	  if((pLastCtrlEvent->ID == TRACE_SYSCALL_EXIT)
	   ||(pLastCtrlEvent->ID == TRACE_TRAP_EXIT)
	   ||(pLastCtrlEvent->ID == TRACE_IRQ_EXIT)
	   ||(pLastCtrlEvent->ID == TRACE_SCHEDCHANGE)
	   ||(pLastCtrlEvent->ID == TRACE_KERNEL_TIMER)
	   ||(pLastCtrlEvent->ID == TRACE_BOTTOM_HALF))
	    {
	    /* Accumulate the time that passed since the last control event */
	    DBTimeSub(lTime, pEvent->Time, pLastCtrlEvent->Time);
	    DBTimeAdd(pProcess->TimeRuning, pProcess->TimeRuning, lTime);

	    /* This is a control event */
	    pEvent->Control = TRUE;
	    pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
	    }
	  }
#if 0
	else
	  {
	  /* This is a control event */
	  pEvent->Control = TRUE;
	  pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
	  }
#endif
#endif
	break;

      /* Exit from a trap */
      case TRACE_TRAP_EXIT :
	pmSystem->TrapExit++;
	break;

      /* Entry into an IRQ */
      case TRACE_IRQ_ENTRY :
	pmSystem->IRQEntry++;

#if 0
	/* Is this a control event */
	if((pEvent->Prev != NULL)
	 &&(pEvent->Prev->Control == TRUE)
	 &&(pProcess->PID != 0)) 
	  {
	  /* Is this the first control event */
	  if(pProcess->LastCtrlEvent != NULL)
	    {
	    /* Was the last control event other than an exit */
	    if((pProcess->LastCtrlEvent->ID == TRACE_SYSCALL_ENTRY)
	     ||(pProcess->LastCtrlEvent->ID == TRACE_TRAP_ENTRY)
	     ||(pProcess->LastCtrlEvent->ID == TRACE_IRQ_ENTRY))
	      /* We entered something while we were already in something else */
	      pProcess->CtrlDepth++;
	    else
	      {
	      /* Accumulate the time that passed since the last control event */
	      DBTimeSub(lTime, pEvent->Time, pProcess->LastCtrlEvent->Time);
	      DBTimeAdd(pProcess->TimeRuning, pProcess->TimeRuning, lTime);
	      }
	    }

	  /* System call entries are always a control event */
	  pEvent->Control = TRUE;
	  pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
	  }
#else
	/* Is this the first control event */
	if((pLastCtrlEvent != NULL)
	 &&(pProcess->PID != 0)) 
	  {
	  /* Was the last control event other than an exit */
	  if((pLastCtrlEvent->ID == TRACE_SYSCALL_EXIT)
	   ||(pLastCtrlEvent->ID == TRACE_TRAP_EXIT)
	   ||(pLastCtrlEvent->ID == TRACE_IRQ_EXIT)
	   ||(pLastCtrlEvent->ID == TRACE_SCHEDCHANGE)
	   ||(pLastCtrlEvent->ID == TRACE_KERNEL_TIMER)
	   ||(pLastCtrlEvent->ID == TRACE_BOTTOM_HALF))
	    {
	    /* Accumulate the time that passed since the last control event */
	    DBTimeSub(lTime, pEvent->Time, pLastCtrlEvent->Time);
	    DBTimeAdd(pProcess->TimeRuning, pProcess->TimeRuning, lTime);

	    /* This is a control event */
	    pEvent->Control = TRUE;
	    pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
	    }
	  }
#if 0
	else
	  {
	  /* This is a control event */
	  pEvent->Control = TRUE;
	  pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
	  }
#endif
#endif
	break;

      /* Exit from an IRQ */
      case TRACE_IRQ_EXIT :
	pmSystem->IRQExit++;
	break;

      /* Task scheduling change */
      case TRACE_SCHEDCHANGE :
	/* Read PID of incoming task */
	lOPID = SCHED_EVENT(pEvent->Struct)->out;
	lPID  = SCHED_EVENT(pEvent->Struct)->in;

	/* Find the process in the list of already known processes */
	for(pProcess = pmSystem->Processes; pProcess != NULL; pProcess = pProcess->Next)
	  if(pProcess->PID == lPID)
	    break;
	
	/* Did we find anything */
	if(pProcess == NULL)
	  /* Create a new process */
	  pProcess = DBCreateProcess(pmSystem, lPID, -1);

	/* Set the owner of the event as being the new incoming task */
	pEvent->Process = pProcess;

	/* Update the number of scheduling changes */
	pmSystem->SchedChange++;

	/* Is this a control event */
	if((lPID != 0)
         &&(pEvent->Next != NULL)
	 &&((pEvent->Next->ID == TRACE_SYSCALL_ENTRY)
	  ||(pEvent->Next->ID == TRACE_TRAP_ENTRY)
	  ||(pEvent->Next->ID == TRACE_IRQ_ENTRY)))
	  {
	  pEvent->Control = TRUE;
	  pProcess->LastCtrlEvent = pLastCtrlEvent = pEvent;
	  }
	break;

      /* The kernel timer routine has been called */
      case TRACE_KERNEL_TIMER :
	pmSystem->KernelTimer++;
	break;

      /* Going to run a bottom half */
      case TRACE_BOTTOM_HALF :
	pmSystem->BH++;
	break;

      /* Hit key part of process management */
      case TRACE_PROCESS :
	/* Don't analyze anything but forking for now */
	if(PROC_EVENT(pEvent->Struct)->event_sub_id != TRACE_PROCESS_FORK)
	  continue;

	/* We have forked, get the child's PID */
	lPID = PROC_EVENT(pEvent->Struct)->event_data1;

	/* Create a new process */
#if 0
	pProcess = DBCreateProcess(pmSystem, lPID, pProcess->PID);
#else       /* When a fork happens we aren't yet runing the created process */
	DBCreateProcess(pmSystem, lPID, pProcess->PID);
#endif
	break;

      /* Hit key part of file system */
      case TRACE_FILE_SYSTEM :
	/* Depending of the event subtype */
	switch(FS_EVENT(pEvent->Struct)->event_sub_id)
	  {
	  /* TRACE_FILE_SYSTEM_BUF_WAIT_START */
	  case TRACE_FILE_SYSTEM_BUF_WAIT_START :
	    /* If we hadn't already started an I/O wait, then we're starting a new one now */
	    if((pProcess->LastIOEvent == NULL)
	     ||((pProcess->LastIOEvent != NULL)
	      &&(FS_EVENT(pProcess->LastIOEvent->Struct)->event_sub_id != TRACE_FILE_SYSTEM_BUF_WAIT_START)))
	      pProcess->LastIOEvent = pEvent;
	    break;

	  /* TRACE_FILE_SYSTEM_BUF_WAIT_END */
	  case TRACE_FILE_SYSTEM_BUF_WAIT_END :
 	    /* We've been waiting in I/O since the last wait start */
	    if((pProcess->LastIOEvent != NULL)
             &&(FS_EVENT(pProcess->LastIOEvent->Struct)->event_sub_id == TRACE_FILE_SYSTEM_BUF_WAIT_START))
	      {
	      DBTimeSub(lTime, pEvent->Time, pProcess->LastIOEvent->Time);
	      DBTimeAdd(pProcess->TimeIO, pProcess->TimeIO, lTime);
	      pProcess->LastIOEvent = pEvent;
	      }
	    break;

	  /* TRACE_FILE_SYSTEM_EXEC */
	  case TRACE_FILE_SYSTEM_EXEC :
	    if(pProcess->Command != NULL) /* Retain only the first exec */
	      continue;
	    pProcess->Command = (char*) malloc(strlen(FS_EVENT(pEvent->Struct)->file_name) + 1);
	    strcpy(pProcess->Command, FS_EVENT(pEvent->Struct)->file_name);
	    break;

	  /* TRACE_FILE_SYSTEM_READ */
	  case TRACE_FILE_SYSTEM_READ :
	    pProcess->QFileRead += FS_EVENT(pEvent->Struct)->event_data2;
	    break;

	  /* TRACE_FILE_SYSTEM_WRITE */
	  case TRACE_FILE_SYSTEM_WRITE :
	    pProcess->QFileWrite += FS_EVENT(pEvent->Struct)->event_data2;
	    break;
	  }
	break;

      /* Hit key part of timer management */
      case TRACE_TIMER :
	/* Did a timer expire */
	if(TIMER_EVENT(pEvent->Struct)->event_sub_id == TRACE_TIMER_EXPIRED)
	  pmSystem->TimerExpire++;
	break;

      /* Hit key part of memory management */
      case TRACE_MEMORY :
	/* Depending on the event sub id */
	switch(MEM_EVENT(pEvent->Struct)->event_sub_id)
	  {
	  /* TRACE_MEMORY_PAGE_ALLOC */
	  case TRACE_MEMORY_PAGE_ALLOC :
	    pmSystem->PageAlloc++;
	    break;

	  /* TRACE_MEMORY_PAGE_FREE */
	  case TRACE_MEMORY_PAGE_FREE :
	    pmSystem->PageFree++;
	    break;
	  }
	break;

      /* Hit key part of inter-process communication */
      case TRACE_IPC :
	break;

      /* Hit key part of network communication */
      case TRACE_NETWORK :
	if(NET_EVENT(pEvent->Struct)->event_sub_id == TRACE_NETWORK_PACKET_OUT)
	  pmSystem->PacketOut++;
	else
	  pmSystem->PacketIn++;
	break;

      default :
	printf("Internal error : Unknow event while processing trace \n");
	exit(1);
	break;
      }
    }

  /* Sum up syscall time for all the system */
  for(pProcess = pmSystem->Processes; pProcess != NULL; pProcess = pProcess->Next)
    /* Go through his syscall list */
    for(pSyscall = pProcess->Syscalls; pSyscall != NULL; pSyscall = pSyscall->Next)
      {
      /* Find this syscall in the system-wide list */
      for(pSyscallP = pmSystem->Syscalls; pSyscallP != NULL; pSyscallP = pSyscallP->Next)
	if(pSyscallP->ID == pSyscall->ID)
	  break;

      /* Did we find anything */
      if(pSyscallP == NULL)
	{
	/* Allocate space for new syscall */
	if((pSyscallP = (syscall*) malloc(sizeof(syscall))) == NULL)
	  {
	  printf("Memory allocation problem \n");
	  exit(1);
	  }

	/* Set and link new syscall */
	pSyscallP->ID    = pSyscall->ID;
	pSyscallP->Nb    = 0;
	pSyscallP->Depth = 0;
	memset(&(pSyscallP->Time), 0, sizeof(struct timeval));
	pSyscallP->Next = pmSystem->Syscalls;
	pmSystem->Syscalls = pSyscallP;
	}

      /* Set number of times system call was called */
      pSyscallP->Nb += pSyscall->Nb;
      
      /* Add time spent in this syscall */
      pSyscallP->Time.tv_sec += pSyscall->Time.tv_sec;
      pSyscallP->Time.tv_usec += pSyscall->Time.tv_usec;

      /* Have we more than a million microseconds */
      if(pSyscallP->Time.tv_usec > 1000000)
	{
	pSyscallP->Time.tv_sec += pSyscallP->Time.tv_usec / 1000000;
	pSyscallP->Time.tv_usec = pSyscallP->Time.tv_usec % 1000000;
	}
      }

  /* Build process tree */
  DBBuildProcTree(pmSystem);
}

/******************************************************************
 * Function :
 *    DBPrintTrace()
 * Description :
 *    Print out the data-base and add details requested at the command-line
 * Parameters :
 *    pDecodedDataFile, File to which the data is to be printed.
 *    pmSystem, Description of the system.
 *    pmTraceDB, The event database to be processed.
 *    pmOptions, The user given options.
 * Return values :
 *    NONE
 * History :
 *    K.Y., 20/05/99, Initial typing.
 * Note :
 ******************************************************************/
void DBPrintTrace(int pmDecodedDataFile, system* pmSystem, db* pmTraceDB, options* pmOptions)
{
  int            lPID;                 /* PID of process to which event belongs */
  long           lDSec, lDUsec;        /* The duration of the trace */
  char           lWriteString[100];    /* String to be written to output file */
  char           lTabing[10];          /* Tab distance between syscall name and time */
  event*         pEvent;               /* Generic event pointer */
  syscall*       pSyscall;             /* Generic syscall pointer */
  process*       pProcess;             /* Generic process poniter */
  struct timeval lTraceDuration;       /* The time the trace lasted */
  struct timeval lTimeProcesses={0,0}; /* Total amount of time during which processes had CPU control */
  struct timeval lTimeSystem={0,0};    /* Total amount of time during which the system had CPU control */

  /* Print some information about the trace  */
  /*  Start time*/
  PrintData(pmDecodedDataFile,
	    "Trace start time: (%ld, %ld) \n",
	    pmTraceDB->StartTime.tv_sec,
	    pmTraceDB->StartTime.tv_usec);
  /*  End time */
  PrintData(pmDecodedDataFile,
	    "Trace end time: (%ld, %ld) \n",
	    pmTraceDB->EndTime.tv_sec,
	    pmTraceDB->EndTime.tv_usec);
  /*  Duration */
  DBTimeSub(lTraceDuration, pmTraceDB->EndTime, pmTraceDB->StartTime);
  lDSec  = lTraceDuration.tv_sec;
  lDUsec = lTraceDuration.tv_usec;
#if 0
  lDSec = pmTraceDB->EndTime.tv_sec - pmTraceDB->StartTime.tv_sec;
  lDUsec = pmTraceDB->EndTime.tv_usec - pmTraceDB->StartTime.tv_usec;
  if(lDUsec < 0)
    {
    lDSec--;
    lDUsec = 1000000 + lDUsec;
    }
#endif
  PrintData(pmDecodedDataFile,
	    "Trace duration: (%ld, %ld) \n\n",
	    lDSec,
	    lDUsec);
  /*  Number of occurences of some events */
  PrintData(pmDecodedDataFile,
	    "Number of occurences of:\n");
  PrintData(pmDecodedDataFile,
	    "\tEvents: %ld \n",
	    pmTraceDB->Nb);
  PrintData(pmDecodedDataFile,
	    "\tScheduling changes: %ld \n",
	    pmSystem->SchedChange);
  PrintData(pmDecodedDataFile,
	    "\tKernel timer tics: %ld \n",
	    pmSystem->KernelTimer);
  PrintData(pmDecodedDataFile,
	    "\tSystem call entries: %ld \n",
	    pmSystem->SyscallEntry);
  PrintData(pmDecodedDataFile,
	    "\tSystem call exits: %ld \n",
	    pmSystem->SyscallExit);
  PrintData(pmDecodedDataFile,
	    "\tTrap entries: %ld \n",
	    pmSystem->TrapEntry);
  PrintData(pmDecodedDataFile,
	    "\tTrap exits: %ld \n",
	    pmSystem->TrapExit);
  PrintData(pmDecodedDataFile,
	    "\tIRQ entries: %ld \n",
	    pmSystem->IRQEntry);
  PrintData(pmDecodedDataFile,
	    "\tIRQ exits: %ld \n",
	    pmSystem->IRQExit);
  PrintData(pmDecodedDataFile,
	    "\tBottom halves: %ld \n",
	    pmSystem->BH);
  PrintData(pmDecodedDataFile,
	    "\tTimer expiries: %ld \n",
	    pmSystem->TimerExpire);
  PrintData(pmDecodedDataFile,
	    "\tPage allocations: %ld \n",
	    pmSystem->PageAlloc);
  PrintData(pmDecodedDataFile,
	    "\tPage frees: %ld \n",
	    pmSystem->PageFree);
  PrintData(pmDecodedDataFile,
	    "\tPackets Out: %ld \n",
	    pmSystem->PacketOut);
  PrintData(pmDecodedDataFile,
	    "\tPackets In: %ld \n\n",
	    pmSystem->PacketIn);

  /* Has the user requested to know the time spent in system calls */
  if(pmOptions->AcctSysCall)
    {
    /* Print summary header */
    PrintData(pmDecodedDataFile,
	      "Analysis details: \n");

    /* Go through the process list */
    for(pProcess = pmSystem->Processes; pProcess != NULL; pProcess = pProcess->Next)
      {
      /* Print basic process information */
      PrintData(pmDecodedDataFile,
		"\tProcess (%d, %d)",
		pProcess->PID,
		pProcess->PPID);
      if(pProcess->Command != NULL)
	{
	PrintData(pmDecodedDataFile,
		  ": %s: \n",
		  pProcess->Command);
	}
      else
	{
	PrintData(pmDecodedDataFile,
		  ": \n");
	}

      /* Print detailed process information */
      PrintData(pmDecodedDataFile,
		"\t\tNumber of system calls: %ld \n",
		pProcess->NbSyscalls);
      PrintData(pmDecodedDataFile,
		"\t\tNumber of traps: %ld \n",
		pProcess->NbTraps);
      PrintData(pmDecodedDataFile,
		"\t\tQuantity of data read from files: %ld \n",
		pProcess->QFileRead);
      PrintData(pmDecodedDataFile,
		"\t\tQuantity of data written to files: %ld \n",
		pProcess->QFileWrite);
      PrintData(pmDecodedDataFile,
		"\t\tTime runing: (%ld, %ld) => %2.2f % \n",
		pProcess->TimeRuning.tv_sec,
		pProcess->TimeRuning.tv_usec,
		(float) 100 * (DBGetTimeInDouble(pProcess->TimeRuning) / DBGetTimeInDouble(lTraceDuration)));
      PrintData(pmDecodedDataFile,
		"\t\tTime waiting for I/O: (%ld, %ld) => %2.2f % \n",
		pProcess->TimeIO.tv_sec,
		pProcess->TimeIO.tv_usec,
		(float) 100 * (DBGetTimeInDouble(pProcess->TimeIO) / DBGetTimeInDouble(lTraceDuration)));

      /* Add the time of this process to the amount of time the CPU was controled by a process */
      DBTimeAdd(lTimeProcesses, lTimeProcesses, pProcess->TimeRuning);

      /* Print out system call detailed information */
      PrintData(pmDecodedDataFile,
		"\t\tSystem call usage: \n");

      /* Go through the list of system calls called by this process */
      for(pSyscall = pProcess->Syscalls; pSyscall != NULL; pSyscall = pSyscall->Next)
	{
	/* Compute tabbing according to syscall's name length */
	if(strlen(SyscallID[pSyscall->ID]) < 6)
	  sprintf(lTabing, "\t\t");
	else if (strlen(SyscallID[pSyscall->ID]) < 13)
	  sprintf(lTabing, "\t");
	else
	  lTabing[0] = '\0';
	
	/* Print system call information */
	PrintData(pmDecodedDataFile,
		  "\t\t%s: %s %d \t (%ld, %ld)\n",
		  SyscallID[pSyscall->ID],
		  lTabing,
		  pSyscall->Nb,
		  pSyscall->Time.tv_sec,
		  pSyscall->Time.tv_usec);
	}
      }

    /* System wide information */
    PrintData(pmDecodedDataFile,
	      "\tSystem-wide:\n");

    /* Print out the total amount of time processes had the control of the cpu */
    PrintData(pmDecodedDataFile,
	      "\t\tTotal process time: (%ld, %ld) => %2.2f % \n",
	      lTimeProcesses.tv_sec,
	      lTimeProcesses.tv_usec,
	      (float) 100 * (DBGetTimeInDouble(lTimeProcesses) / DBGetTimeInDouble(lTraceDuration)));

    /* Print out the total amount of time linux had the control of the cpu */
    DBTimeSub(lTimeSystem, lTraceDuration, lTimeProcesses)
    PrintData(pmDecodedDataFile,
	      "\t\tTotal system time: (%ld, %ld) => %2.2f % \n",
	      lTimeSystem.tv_sec,
	      lTimeSystem.tv_usec,
	      (float) 100 * (DBGetTimeInDouble(lTimeSystem) / DBGetTimeInDouble(lTraceDuration)));

    /* Print out system call detailed information */
    PrintData(pmDecodedDataFile,
	      "\t\tSystem call usage: \n");

    /* Go through the system-wide syscall accouting list */
    for(pSyscall = pmSystem->Syscalls; pSyscall != NULL; pSyscall = pSyscall->Next)
      {
      /* Compute tabbing according to syscall's name length */
      if(strlen(SyscallID[pSyscall->ID]) < 6)
	sprintf(lTabing, "\t\t");
      else if (strlen(SyscallID[pSyscall->ID]) < 13)
	sprintf(lTabing, "\t");
      else
	lTabing[0] = '\0';
      
      /* Print system call information */
      PrintData(pmDecodedDataFile,
		"\t\t%s: %s %d \t (%ld, %ld)\n",
		SyscallID[pSyscall->ID],
		lTabing,
		pSyscall->Nb,
		pSyscall->Time.tv_sec,
		pSyscall->Time.tv_usec);
      }
    }

  PrintData(pmDecodedDataFile,
	    "\n");

  /* Are we tracing only one CPUID */
  if(pmOptions->TraceCPUID == TRUE)
    PrintData(pmDecodedDataFile,
	      "Tracing CPU %d only \n\n",
	      pmOptions->CPUID);

  /* Are we tracing only one PID */
  if(pmOptions->TracePID == TRUE)
    PrintData(pmDecodedDataFile,
	      "Tracing process %d only \n\n",
	      pmOptions->PID);

  /* Was it only a summary that we wanted */
  if(pmOptions->Summarize == TRUE)
    return;

  /* Print the output file's event header */
  PrintHeader(pmDecodedDataFile, &(pmTraceDB->Properties), pmOptions);

  /* Go through the events list */
  for(pEvent = pmTraceDB->Events; pEvent != NULL; pEvent = pEvent->Next)
    {
    /* Does this event belong to anyone */
    if((pEvent->Process == NULL) && (pmOptions->TracePID == TRUE))
      continue;

    /* Did the user request this event to be printed */
    if((((pmOptions->Omit == TRUE) && (test_bit(pEvent->ID, &(pmOptions->OmitMask))))
       ||((pmOptions->Trace == TRUE) && (test_bit(pEvent->ID, &(pmOptions->TraceMask))))
       ||((pmOptions->Trace != TRUE) && (pmOptions->Omit != TRUE)))
     &&(((pmTraceDB->Properties.LogCPUID == TRUE)
        &&(((pmOptions->TraceCPUID == TRUE) && (pmOptions->CPUID == pEvent->CPUID))
          ||(pmOptions->TraceCPUID != TRUE)))
       ||(pmTraceDB->Properties.LogCPUID != TRUE))
     &&(((pmOptions->TracePID == TRUE) && (pmOptions->PID == pEvent->Process->PID))
       ||(pmOptions->TracePID != TRUE)
       ||((pmOptions->TracePID == TRUE)
         &&(pEvent->Prev != NULL)
         &&(pEvent->Prev->Process != NULL)
         &&(pEvent->ID == TRACE_SCHEDCHANGE)
         &&(pEvent->Prev->Process->PID == pmOptions->PID))))
      {
      /* Does this process belong to anyone */
      if(pEvent->Process == NULL)
	lPID = -1;
      else
	lPID = pEvent->Process->PID;

      /* Output this event */
      PrintEvent(pmDecodedDataFile,
		 pEvent->CPUID,
		 pEvent->ID,
		 &(pEvent->Time),
		 lPID,
		 pEvent->Size,
		 pEvent->String,
		 &(pmTraceDB->Properties),
		 pmOptions);
      }
    }
}

/******************************************************************
 * Function :
 *    DBCreateProcess()
 * Description :
 *    Creates a process, initializes it, inserts it into the list
 *    of processes belonging to the given system and returns it
 *    to the caller.
 * Parameters :
 *    pmSystem, the system to which the process belongs.
 *    pmPID, the PID of the process being created.
 *    pmPPID, the PID of the process' parent.
 * Return values :
 *    The newly created process.
 * History :
 *    K.Y. 03/11/99. Intial typing.
 ******************************************************************/
process* DBCreateProcess(system* pmSystem, int pmPID, int pmPPID)
{
  process*  pProcess;             /* Generic process pointer */
  
  /* Allocate space for a new process */
  if((pProcess = (process*) malloc(sizeof(process))) == NULL)
    {
    printf("Memory allocation problem \n");
    exit(1);
    }
	  
  /* Set new process's properties */
  pProcess->PID        = pmPID;
  pProcess->PPID       = pmPPID;
  pProcess->Command    = NULL;
  pProcess->NbSyscalls = 0;
  pProcess->NbTraps    = 0;
  pProcess->QFileRead  = 0;
  pProcess->QFileWrite = 0;

  pProcess->TimeRuning.tv_sec  = 0;
  pProcess->TimeRuning.tv_usec = 0;
  pProcess->TimeIO.tv_sec      = 0;
  pProcess->TimeIO.tv_usec     = 0;

  pProcess->Depth         = 0;
  pProcess->Pending       = NULL;
  pProcess->CtrlDepth     = 0;
  pProcess->LastCtrlEvent = NULL;
  pProcess->LastIOEvent   = NULL;

  pProcess->Syscalls   = NULL;
  pProcess->System     = pmSystem;

  pProcess->Next       = NULL;
  pProcess->Children   = NULL;
  pProcess->NextChild  = NULL;

  /* Insert new process in processes list */
  pProcess->Next = pmSystem->Processes;
  pmSystem->Processes = pProcess;    

  /* Give the newly created process to the caller */
  return pProcess;
}

/******************************************************************
 * Function :
 *    DBCreateDB()
 * Description :
 *    Creates a database of events, initializes it and returns it
 *    to the caller.
 * Parameters :
 *    None.
 * Return values :
 *    The newly created database.
 * History :
 *    K.Y. 21/10/99. Intial typing.
 ******************************************************************/
db* DBCreateDB(void)
{
  db*   pDB;   /* The newly create database */

  /* Allocate space for database */
  if((pDB = (db*) malloc(sizeof(db))) == NULL)
    {
    printf("Memory allocation problem \n");
    exit(1);
    }

  /* Initialize database */
  pDB->Nb        = 0;
  pDB->Events    = NULL;
  pDB->LastEvent = NULL;

  /* Give the new database to the caller */
  return pDB;
}

/******************************************************************
 * Function :
 *    DBCreateSystem()
 * Description :
 *    Creates a system, initializes it and returns it to the caller.
 * Parameters :
 *    None.
 * Return values :
 *    The newly created system.
 * History :
 *    K.Y. 21/10/99. Intial typing.
 ******************************************************************/
system* DBCreateSystem(void)
{
  system*  pSystem;  /* The newly created system */

  /* Allocate space for database */
  if((pSystem = (system*) malloc(sizeof(system))) == NULL)
    {
    printf("Memory allocation problem \n");
    exit(1);
    }

  /* Initialize the system */
  pSystem->Processes  = NULL;
  pSystem->ProcTree   = NULL;
  pSystem->Syscalls   = NULL;

  /* Give the new system to the caller */
  return pSystem;
}

/******************************************************************
 * Function :
 *    DBDestroyTrace()
 * Description :
 * Parameters :
 * Return values :
 * History :
 * Note :
 ******************************************************************/
void DBDestroyTrace(system* pmSystem, db* pmEventDB)
{
  event*    pEvent;       /* Generic event pointer */
  event*    pEventTemp;   /* Temporary event pointer */
  syscall*  pSysCall;     /* Generic syscall pointer */
  syscall*  pSysCallTemp; /* Temporary syscall pointer */
  process*  pProc;        /* Generic process pointer */
  process*  pProcTemp;    /* Temporary process pointer */

  /* Go through event database */
  for(pEvent = pmEventDB->Events; pEvent != NULL;)
    {
    /* Remember next event in list */
    pEventTemp = pEvent->Next;

    /* Free memory used by event string, if any */
    if(pEvent->String != NULL)
      free(pEvent->String);

    /* Free memory used by event */
    free(pEvent);

    /* Go to next event */
    pEvent = pEventTemp;
    }

  /* Free memory used by event database */
  free(pmEventDB);

  /* Go through process list */
  for(pProc = pmSystem->Processes; pProc != NULL;)
    {
    /* Remember next process in list */
    pProcTemp = pProc->Next;

    /* Free memory used by command, if any */
    if(pProc->Command != NULL)
      free(pProc->Command);

    /* Go through list of syscalls for this process */
    for(pSysCall = pProc->Syscalls; pSysCall != NULL;)
      {
      /* Remember next syscall in syscall list */
      pSysCallTemp = pSysCall->Next;

      /* Free space taken by syscall */
      free(pSysCall);

      /* Go to next syscall */
      pSysCall = pSysCallTemp;
      }

    /* Go through list of pending  syscalls for this process (this is PARANOIA) */
    for(pSysCall = pProc->Pending; pSysCall != NULL;)
      {
      /* Remember next syscall in syscall list */
      pSysCallTemp = pSysCall->Next;

      /* Free space taken by syscall */
      free(pSysCall);

      /* Go to next syscall */
      pSysCall = pSysCallTemp;
      }

    /* Free memory used by process */
    free(pProc);

    /* Go to next process */
    pProc = pProcTemp;
    }

  /* Go through list of system-wide system call accouting */
  for(pSysCall = pmSystem->Syscalls; pSysCall != NULL;)
    {
    /* Remember next syscall in syscall list */
    pSysCallTemp = pSysCall->Next;

    /* Free space taken by syscall */
    free(pSysCall);
    
    /* Go to next syscall */
    pSysCall = pSysCallTemp;
    }

  /* Free memory used by system */
  free(pmSystem);
}
