/*
   FileIO.c : File I/O engine.
   Copyright (C) 1999 Karim Yaghmour.
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   History : 
     K.Y., 11/10/99, Modified to conform with complete kernel instrumentation
     K.Y., 26/06/99, Separeted engine from main project file.
*/

#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "TraceToolkit.h"
#include "Tables.h"
#include "FileIO.h"

/***************************************************************************************/
/************************************ Functions ****************************************/
/***************************************************************************************/
/******************************************************************
 * Function :
 *    ReadRawEntry()
 * Description :
 *    Reads a raw data entry from a trace file and fills the
 *    given fields.
 * Parameters :
 *    pmTraceDataFile, File from which data is read.
 *    pmProperties, Properties of the read trace.
 *    pmEventID, Identity of event.
 *    pmEventTime, Time at which event occured. The caller must pass
 *                 the time of the previous event and he receives
 *                 the time of the current event.
 *    pmEventCPUID, CPU on which event took place.
 *    pmEventStruct, Structure describing event.
 *    pmEventString, String describing the event.
 *    pmEventDataSize, Size of event data.
 * Return values :
 *    TRUE, Entry read OK.
 *    FALSE, We reached EOF or trace is invalid
 * History :
 *    K.Y., 11/10/99, Changed call prototype.
 *    K.Y., 09/09/99, Modified for compressed data.
 *    K.Y., 20/05/99, Initial typing.
 * Note :
 ******************************************************************/
int ReadRawEntry(int              pmTraceDataFile,
		 traceProperties* pmProperties,
		 int*             pmEventID,
		 struct timeval*  pmEventTime,
		 int*             pmEventCPUID,
		 void**           pmEventStruct,
		 char**           pmEventString,
		 int*             pmEventDataSize)
{
  int               lSizeRead;              /* Quantity of data read from file */
  int               lEventDataSize;         /* Size of event data */
  int               lEventStringSize;       /* Size of event string */
  char              lEventStringTemp[512];  /* Temporary string used to build string describing event */   
  char*             lEventString = NULL;    /* String describing the event */
  char*             lFileName;              /* String holding a filename */
  void*             pEventStruct = NULL;    /* Structure containing event details */
  int               lEventCPUID = -1;       /* The ID of the CPU where the event took place */
  uint8_t           lCPUIDTemp;             /* Temporary holding place for CPUID */
  uint8_t           lEventID;               /* Identy of event */
  uint16_t          lEventTimeDelta;        /* Time elapsed between current and last event */

  /* Initialize event string */
  lEventStringTemp[0] = '\0';

  /* Initialize return values */
  *pmEventDataSize = 0;

  /* Read the CPUID, if any */
  if((pmProperties->State == INIT) && (pmProperties->LogCPUID == TRUE))
    {
    /* Read the CPU ID */
    lSizeRead = read(pmTraceDataFile, &lCPUIDTemp, sizeof(lCPUIDTemp));
    if(!lSizeRead)
      return FALSE;
    *pmEventDataSize += sizeof(lEventCPUID);
    lEventCPUID = (int) lCPUIDTemp;
    }

  /* Read event ID */
  lSizeRead = read(pmTraceDataFile, &lEventID, sizeof(lEventID));
  if(!lSizeRead)
    return FALSE;
  *pmEventDataSize += sizeof(lEventID);

  /* Is this a know event */
  if(lEventID > TRACE_MAX)
    return FALSE;

  /* Read event time delta */
  lSizeRead = read(pmTraceDataFile, &lEventTimeDelta, sizeof(lEventTimeDelta));
  if(!lSizeRead)
    return FALSE;
  *pmEventDataSize += sizeof(lEventTimeDelta);

  /* Is there a structure describing this event */
  if(((EventStructSize[lEventID] > 0) && (pmProperties->State == INIT))
  || (pmProperties->State == UNINIT))
    {
    if((test_bit(lEventID, &(pmProperties->DetailsMask)) || (pmProperties->State == UNINIT)))
      {
      /* Allocate space for the structure */
      pEventStruct = (void*) malloc(EventStructSize[lEventID]);

      /* Read the structure */
      lSizeRead = read(pmTraceDataFile, pEventStruct, EventStructSize[lEventID]);
      if(!lSizeRead)
	{
	free(pEventStruct);
	return FALSE;
	}
      
      /* Update event data size */
      *pmEventDataSize += EventStructSize[lEventID];
      }
    else
      goto final_setting; /* I use a goto only because it makes the code more readable. */
    }

  /* Depending on the event's type */
  switch(lEventID)
    {
    /* TRACE_START */
    case TRACE_START :
      /* Does this trace fit */
      if((START_EVENT(pEventStruct)->MagicNumber != TRACER_MAGIC_NUMBER)
      || (START_EVENT(pEventStruct)->MajorVersion < TRACER_SUP_VERSION_MAJOR)
      || ((START_EVENT(pEventStruct)->MajorVersion >= TRACER_SUP_VERSION_MAJOR)
       && (START_EVENT(pEventStruct)->MinorVersion < TRACER_SUP_VERSION_MINOR)))
	{
	/* This isn't a valid trace */
	pmProperties->State = INVALID;
	return FALSE;
	}

      /* Set the time of the current event */
      memcpy(pmEventTime,
	     &(START_EVENT(pEventStruct)->Time),
	     sizeof(struct timeval));

      /* Set the properties of the trace */
      pmProperties->State       = INIT;
      pmProperties->EventMask   = START_EVENT(pEventStruct)->EventMask;
      pmProperties->DetailsMask = START_EVENT(pEventStruct)->DetailsMask;
      pmProperties->LogCPUID    = START_EVENT(pEventStruct)->LogCPUID;

      /* Format the string describing the event */
      sprintf(lEventStringTemp,
	      "START AT : (%ld, %ld)",
	      pmEventTime->tv_sec,
	      pmEventTime->tv_usec);
      break;

    /* TRACE_SYSCALL_ENTRY */
    case TRACE_SYSCALL_ENTRY :
      /* Format the string describing the event */
      sprintf(lEventStringTemp,
	      "SYSCALL : %s; EIP : 0x%8X",
	      SyscallID[SYSCALL_EVENT(pEventStruct)->syscall_id],
	      SYSCALL_EVENT(pEventStruct)->address);
      break;

    /* TRACE_TRAP_ENTRY */
    case TRACE_TRAP_ENTRY :
      /* Format the string describing the event */
      sprintf(lEventStringTemp,
	      "TRAP : %s; EIP : 0x%8X",
	      TrapID[TRAP_EVENT(pEventStruct)->trap_id],
	      TRAP_EVENT(pEventStruct)->address);
      break;

    /* TRACE_IRQ_ENTRY */
    case TRACE_IRQ_ENTRY :
      /* Format the string describing the event */
      sprintf(lEventStringTemp,
	      "IRQ : %d",
	      *IRQ_EVENT(pEventStruct));
      break;

    /* TRACE_SCHEDCHANGE */
    case TRACE_SCHEDCHANGE :
      /* Format the string describing the event */
      sprintf(lEventStringTemp,
	      "IN : %d; OUT : %d; STATE : %d",
	      SCHED_EVENT(pEventStruct)->in,
	      SCHED_EVENT(pEventStruct)->out,
	      SCHED_EVENT(pEventStruct)->out_state);
      break;

    /* TRACE_BOTTOM_HALF */
    case TRACE_BOTTOM_HALF :
      /* Format the string describing the event */
      sprintf(lEventStringTemp,
	      "BH : %d",
	      *BH_EVENT(pEventStruct));
      break;

    /* TRACE_PROCESS */
    case TRACE_PROCESS :
      /* Depending on the process event */
      switch(PROC_EVENT(pEventStruct)->event_sub_id)
	{
	/* TRACE_PROCESS_KTHREAD */
	case TRACE_PROCESS_KTHREAD :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "KTHREAD : %d; FUNCTION ADDRESS : 0x%8X",
		  PROC_EVENT(pEventStruct)->event_data1,
		  PROC_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_PROCESS_FORK */
	case TRACE_PROCESS_FORK :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "FORK : %d",
		  PROC_EVENT(pEventStruct)->event_data1);
	  break;

	/* TRACE_PROCESS_EXIT */
	case TRACE_PROCESS_EXIT :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "EXIT");
	  break;

	/* TRACE_PROCESS_WAIT */
	case TRACE_PROCESS_WAIT :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "WAIT PID : %d" /* The PID waited for */,
		  PROC_EVENT(pEventStruct)->event_data1);
	  break;

	/* TRACE_PROCESS_SIGNAL */
	case TRACE_PROCESS_SIGNAL :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SIGNAL : %d; PID : %d" /* Signal nbr and destination PID */,
		  PROC_EVENT(pEventStruct)->event_data1,
		  PROC_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_PROCESS_WAKEUP */
	case TRACE_PROCESS_WAKEUP :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "WAKEUP PID : %d; STATE : %d" /* The PID to wake up and it's state before waking */,
		  PROC_EVENT(pEventStruct)->event_data1,
		  PROC_EVENT(pEventStruct)->event_data2);
	  break;
	}
      break;

    /* TRACE_FILE_SYSTEM */
    case TRACE_FILE_SYSTEM :
      /* Depending on the file system event */
      switch(FS_EVENT(pEventStruct)->event_sub_id)
	{
	/* TRACE_FILE_SYSTEM_BUF_WAIT_START */
	case TRACE_FILE_SYSTEM_BUF_WAIT_START :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "START I/O WAIT");
	  break;

	/* TRACE_FILE_SYSTEM_BUF_WAIT_END */
	case TRACE_FILE_SYSTEM_BUF_WAIT_END :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "END I/O WAIT");
	  break;

	/* TRACE_FILE_SYSTEM_EXEC or TRACE_FILE_SYSTEM_OPEN */
	case TRACE_FILE_SYSTEM_EXEC :
	case TRACE_FILE_SYSTEM_OPEN :
	  /* Count the length of the filename in the entry's length */
	  *pmEventDataSize += FS_EVENT(pEventStruct)->event_data2;
	 
	  /* Allocate space to read the filename */
	  if((lFileName = (char*) malloc(FS_EVENT(pEventStruct)->event_data2 + 1)) == NULL)
	    {
	    /* Couldn't get any memory */
	    printf("Memory allocation problem ... Exiting \n");
	    exit(1);
	    }
	  
	  /* Read the filename */ 
	  lSizeRead = read(pmTraceDataFile, lFileName, FS_EVENT(pEventStruct)->event_data2 + 1);
	  if(!lSizeRead)
	    {
	    free(lFileName);
	    free(pEventStruct);
	    return FALSE;
	    }

	  /* Format the string describing the event */
	  if(lEventID == TRACE_FILE_SYSTEM_EXEC)
	    sprintf(lEventStringTemp,
		    "EXEC : %s",
		    lFileName);
	  else
	    sprintf(lEventStringTemp,
		    "OPEN : %s; FD : %d" /* The opened file and the returned file descriptor */,
		    lFileName,
		    FS_EVENT(pEventStruct)->event_data1);

	  /* Keep the filename in the event structure */
	  FS_EVENT(pEventStruct)->file_name = lFileName;
	  break;

	/* TRACE_FILE_SYSTEM_CLOSE */
	case TRACE_FILE_SYSTEM_CLOSE :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "CLOSE : %d",
		  FS_EVENT(pEventStruct)->event_data1);
	  break;

	/* TRACE_FILE_SYSTEM_READ */
	case TRACE_FILE_SYSTEM_READ :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "READ : %d; COUNT : %d" /* FD read and how much */,
		  FS_EVENT(pEventStruct)->event_data1,
		  FS_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_FILE_SYSTEM_WRITE */
	case TRACE_FILE_SYSTEM_WRITE :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "WRITE : %d; COUNT : %d" /* FD written to and how much */,
		  FS_EVENT(pEventStruct)->event_data1,
		  FS_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_FILE_SYSTEM_SEEK */
	case TRACE_FILE_SYSTEM_SEEK :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SEEK : %d; OFFSET : %d",
		  FS_EVENT(pEventStruct)->event_data1,
		  FS_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_FILE_SYSTEM_IOCTL */
	case TRACE_FILE_SYSTEM_IOCTL :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "IOCTL : %d; COMMAND : 0x%X" /* FD ioctled on and command */,
		  FS_EVENT(pEventStruct)->event_data1,
		  FS_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_FILE_SYSTEM_SELECT */
	case TRACE_FILE_SYSTEM_SELECT :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SELECT : %d; TIMEOUT : %ld" /* FD select on and timeout */,
		  FS_EVENT(pEventStruct)->event_data1,
		  (signed long) (FS_EVENT(pEventStruct)->event_data2));
	  break;

	/* TRACE_FILE_SYSTEM_POLL */
	case TRACE_FILE_SYSTEM_POLL :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "POLL : %d; TIMEOUT : %d" /* FD polled on and timeout */,
		  FS_EVENT(pEventStruct)->event_data1,
		  FS_EVENT(pEventStruct)->event_data2);
	  break;
	}
      break;

    /* TRACE_TIMER */
    case TRACE_TIMER :
      /* Depending on the timer event */
      switch(TIMER_EVENT(pEventStruct)->event_sub_id)
	{
	/* TRACE_TIMER_EXPIRED */
	case TRACE_TIMER_EXPIRED :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "TIMER EXPIRED");
	  break;

	/* TRACE_TIMER_SETITIMER */
	case TRACE_TIMER_SETITIMER :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SET ITIMER : %d; INTERVAL : %d; VALUE : %d"  /* WHICH, interval, value */,
		  TIMER_EVENT(pEventStruct)->event_sdata,
		  TIMER_EVENT(pEventStruct)->event_data1,
		  TIMER_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_TIMER_SETTIMEOUT */
	case TRACE_TIMER_SETTIMEOUT :
	  /* Format the string describing the event */
#if 0
	  sprintf(lEventStringTemp,
		  "SET TIMEOUT : %ld",
		  (long) TIMER_EVENT(pEventStruct)->event_data1);
#endif
	  sprintf(lEventStringTemp,
		  "SET TIMEOUT : %d",
		  TIMER_EVENT(pEventStruct)->event_data1);
	  break;
	}
      break;

    /* TRACE_MEMORY */
    case TRACE_MEMORY :
      /* Depending on the memory event */
      switch(MEM_EVENT(pEventStruct)->event_sub_id)
	{
	/* TRACE_MEMORY_PAGE_ALLOC */
	case TRACE_MEMORY_PAGE_ALLOC :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "PAGE ALLOC ORDER : %ld",
		  MEM_EVENT(pEventStruct)->event_data);
	  break;

	/* TRACE_MEMORY_PAGE_FREE */
	case TRACE_MEMORY_PAGE_FREE :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "PAGE FREE ORDER : %ld" /* The adress of what is freed is useless, it's size is */,
		  MEM_EVENT(pEventStruct)->event_data);
	  break;

	/* TRACE_MEMORY_SWAP_IN */
	case TRACE_MEMORY_SWAP_IN :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SWAP IN : 0x%8X" /* Page being swapped in */,
		  MEM_EVENT(pEventStruct)->event_data);
	  break;

	/* TRACE_MEMORY_SWAP_OUT */
	case TRACE_MEMORY_SWAP_OUT :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SWAP OUT : 0x%8X" /* Page being swapped OUT */,
		  MEM_EVENT(pEventStruct)->event_data);
	  break;

	/* TRACE_MEMORY_PAGE_WAIT_START */
	case TRACE_MEMORY_PAGE_WAIT_START :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "START PAGE WAIT");
	  break;

	/* TRACE_MEMORY_PAGE_WAIT_END */
	case TRACE_MEMORY_PAGE_WAIT_END :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "END PAGE WAIT");
	  break;
	}
      break;

    /* TRACE_IPC */
    case TRACE_IPC :
      /* Depending on the ipc event */
      switch(IPC_EVENT(pEventStruct)->event_sub_id)
	{
	/* TRACE_IPC_SOCKET_CALL */
	case TRACE_IPC_SOCKET_CALL :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SO; CALL : %d; FPM(FD) : %d" /* call's id and first parameter */,
		  IPC_EVENT(pEventStruct)->event_data1,
		  IPC_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_IPC_SOCKET_CREATE */
	case TRACE_IPC_SOCKET_CREATE :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SO CREATE; FD : %d; TYPE : %d",
		  IPC_EVENT(pEventStruct)->event_data1,
		  IPC_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_IPC_SOCKET_SEND */
	case TRACE_IPC_SOCKET_SEND :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SO SEND; TYPE : %d; SIZE : %d",
		  IPC_EVENT(pEventStruct)->event_data1,
		  IPC_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_IPC_SOCKET_RECEIVE */
	case TRACE_IPC_SOCKET_RECEIVE :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SO RECEIVE; TYPE : %d; SIZE : %d",
		  IPC_EVENT(pEventStruct)->event_data1,
		  IPC_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_IPC_IPC_CALL */
	case TRACE_IPC_IPC_CALL :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "IPC; CALL : %d; FPM(ID) : %d" /* call's id and first parameter */,
		  IPC_EVENT(pEventStruct)->event_data1,
		  IPC_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_IPC_IPC_MSG_CREATE */
	case TRACE_IPC_MSG_CREATE :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "MSG CREATE; ID : %d; FLAGS : 0x%8X",
		  IPC_EVENT(pEventStruct)->event_data1,
		  IPC_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_IPC_IPC_SEM_CREATE */
	case TRACE_IPC_SEM_CREATE :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SEM CREATE; ID : %d; FLAGS : 0x%8X",
		  IPC_EVENT(pEventStruct)->event_data1,
		  IPC_EVENT(pEventStruct)->event_data2);
	  break;

	/* TRACE_IPC_IPC_SHM_CREATE */
	case TRACE_IPC_SHM_CREATE :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "SHM CREATE; ID : %d; FLAGS : 0x%8X",
		  IPC_EVENT(pEventStruct)->event_data1,
		  IPC_EVENT(pEventStruct)->event_data2);
	  break;
	}
      break;

    /* TRACE_NETWORK */
    case TRACE_NETWORK :
      /* Depending on the network event */
      switch(NET_EVENT(pEventStruct)->event_sub_id)
	{
	/* TRACE_NETWORK_PACKET_IN */
	case TRACE_NETWORK_PACKET_IN :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "PACKET IN; PROTOCOL : %d",
		  NET_EVENT(pEventStruct)->event_data);
	  break;
	/* TRACE_NETWORK_PACKET_OUT */
	case TRACE_NETWORK_PACKET_OUT :
	  /* Format the string describing the event */
	  sprintf(lEventStringTemp,
		  "PACKET OUT; PROTOCOL : %d",
		  NET_EVENT(pEventStruct)->event_data);
	  break;
	}
      break;
    }

  /* Compute event string size */
  lEventStringSize = strlen(lEventStringTemp);

  /* Is there an event description string */
  if(lEventStringSize > 0)
    {
    /* Allocate memory for the event string */
    if((lEventString = (char*) malloc(lEventStringSize + 1)) == NULL)
      {
      /* Couldn't get any memory */
      printf("Memory allocation problem ... Exiting \n");
      exit(1);
      }

    /* Copy string */
    memcpy(lEventString,
	   lEventStringTemp,
	   lEventStringSize);
    lEventString[lEventStringSize] = '\0';
    }

final_setting:

  /* Set the given fields to the values found */
  *pmEventID        = (int) lEventID;
  *pmEventCPUID     = lEventCPUID;
  *pmEventStruct    = pEventStruct;
  *pmEventString    = lEventString;

  /* If this isn't the first event, update event time */
  if(lEventID != TRACE_START)
    {
    /* Increment the number of microseconds */
    pmEventTime->tv_usec += (long) lEventTimeDelta;

    /* Did we go over the million */
    if(pmEventTime->tv_usec > 1000000)
      {
      pmEventTime->tv_sec++;
      pmEventTime->tv_usec -= 1000000;
      }
    }

  /* All OK */
  return TRUE;
}

/******************************************************************
 * Function :
 *    PrintHeader()
 * Description :
 *    Prints the output file header according to the user options.
 * Parameters :
 *    pmDecodedDataFile, File to which the header is written.
 *    pmProperties, The properties of the trace for which the header
 *                  is printed.
 *    pmOptions, The user given options.
 * Return values :
 *    NONE
 * History :
 *    K.Y., 17/10/99, Added support for the trace properties
 *    K.Y., 25/05/99, Initial typing.
 * Note :
 ******************************************************************/
void PrintHeader(int              pmDecodedDataFile,
		 traceProperties* pmProperties,
		 options*         pmOptions)
{
  char              lWriteString[100];      /* String to be written to output file */

  /* Write a header to the output file */
  /*  The baner */
  PrintData(pmDecodedDataFile, HEADER_BANER);

  /*  The CPUID */
  if((pmOptions->ForgetCPUID != TRUE) && (pmProperties->LogCPUID == TRUE))
    PrintData(pmDecodedDataFile, "%s \t", HEADER_CPUID); 

  /*  The Event ID */
  PrintData(pmDecodedDataFile, HEADER_EVENT);

  /*  The time */
  if(pmOptions->ForgetTime != TRUE)
    PrintData(pmDecodedDataFile, HEADER_TIME);

  /*  The PID */
  if(pmOptions->ForgetPID != TRUE)
    PrintData(pmDecodedDataFile, HEADER_PID);

  /*  The data length */
  if(pmOptions->ForgetDataLen != TRUE)
    PrintData(pmDecodedDataFile, HEADER_LENGTH);

  /*  The description string */
  if(pmOptions->ForgetString != TRUE)
    PrintData(pmDecodedDataFile, HEADER_STRING);

  /*  Jump a line */
  PrintData(pmDecodedDataFile, "\n");

  /*  The baner */
  PrintData(pmDecodedDataFile, HEADER_BANER);
}

/******************************************************************
 * Function :
 *    PrintEvent()
 * Description :
 *    Prints an event to the given output file according to the
 *    options.
 * Parameters :
 *    pmDecodedDataFile, File to which the event is print.
 *    pmEventCPUID, The CPU-ID of the cpu on which event took place
 *    pmEventID, The event's ID.
 *    pmEventTime, The time at which the event occured.
 *    pmEventPID, The PID of the process to which this event belongs
 *    pmEventDataSize, The size of the tracing driver entry.
 *    pmEventString, String describing the event.
 *    pmProperties, The properties of the trace to which this event
 *                  belongs
 *    pmOptions, The user given options.
 * Return values :
 *    NONE
 * History :
 *    K.Y., 17/10/99, Added support for the trace properties and
 *                    extend list of parameters.
 *    K.Y., 25/05/99, Initial typing.
 * Note :
 ******************************************************************/
void PrintEvent(int              pmDecodedDataFile,
		int              pmEventCPUID,
		int              pmEventID,
		struct timeval*  pmEventTime,
		int              pmEventPID,
		int              pmEventDataSize,
		char*            pmEventString,
		traceProperties* pmProperties,
		options*         pmOptions)
{
  char              lWriteString[100];      /* String to be written to output file */

  /* Write it all in a readable format according to the user's options */
  /*  The CPUID */
  if((pmOptions->ForgetCPUID != TRUE) && (pmProperties->LogCPUID == TRUE))
    PrintData(pmDecodedDataFile, "%d \t", pmEventCPUID);

  /*  The event ID */
  PrintData(pmDecodedDataFile, "%s \t", EventID[pmEventID]);

  /*  An extra space */
  if((pmEventID == TRACE_TIMER)
   ||(pmEventID == TRACE_MEMORY)
   ||(pmEventID == TRACE_IPC)
   ||(pmEventID == TRACE_NETWORK))
    PrintData(pmDecodedDataFile, "\t");

  /*  The time */
  if(pmOptions->ForgetTime != TRUE)
    PrintData(pmDecodedDataFile, "(%ld, %ld) \t", pmEventTime->tv_sec, pmEventTime->tv_usec);

  /*  The PID */
  if(pmOptions->ForgetPID != TRUE)
    PrintData(pmDecodedDataFile, "%d \t", pmEventPID);

  /*  The data size */
  if(pmOptions->ForgetDataLen != TRUE)
    PrintData(pmDecodedDataFile, "%d \t", pmEventDataSize);

  /*  The description string */
  if((pmEventString != NULL) && (pmOptions->ForgetString != TRUE))
    PrintData(pmDecodedDataFile, "%s", pmEventString);

  /*  A carriage return line-feed */
  PrintData(pmDecodedDataFile, "\n");
}

/******************************************************************
 * Function :
 *    DumpToOutput()
 * Description :
 *    Dumps input as is in human readable format to the output.
 * Parameters :
 *    pmTraceDataFile, File containing the raw trace data.
 *    pmDecodedDataFile, File to which the decoded data is dumped.
 *    pmOptions, The user given options.
 * Return values :
 *    NONE
 * History :
 *    K. Yaghmour, 19/05/99, Initial typing.
 * Note :
 ******************************************************************/
void DumpToOutput(int pmTraceDataFile, int  pmDecodedDataFile, options* pmOptions)
{
  int               lEventID;               /* Identy of event */
  struct timeval    lEventTime;             /* The time at which the event occured */
  int               lEventCPUID;            /* The CPUID */
  void*             lEventStruct;           /* Structure describing the event */
  char*             lEventString;           /* String describing the event */
  int               lEventDataSize;         /* Size of event data */
  traceProperties   lProperties;            /* The properties of the input file */
  int               lCurrentPID = -1;       /* PID of process to which current events belong */

  /* Set properties to unitialized */
  lProperties.State = UNINIT;
  
  /* Read the first event */
  if((ReadRawEntry(pmTraceDataFile,
		   &lProperties,
		   &lEventID,
		   &lEventTime,
		   &lEventCPUID,
		   &lEventStruct,
		   &lEventString,
		   &lEventDataSize) != TRUE) || (lProperties.State != INIT))
    return;

  /* Print the header */
  PrintHeader(pmDecodedDataFile, &lProperties, pmOptions);

  /* Print the first entry */
  PrintEvent(pmDecodedDataFile,
	     lEventCPUID,
	     lEventID,
	     &lEventTime,
	     lCurrentPID,
	     lEventDataSize,
	     lEventString,
	     &lProperties,
	     pmOptions);

  /* Read it through */
  while(ReadRawEntry(pmTraceDataFile,
		     &lProperties,
		     &lEventID,
		     &lEventTime,
		     &lEventCPUID,
		     &lEventStruct,
		     &lEventString,
		     &lEventDataSize) == TRUE)
    {
    /* Is this a scheduling change */
    if(lEventID == TRACE_SCHEDCHANGE)
      lCurrentPID = SCHED_EVENT(lEventStruct)->in;

    /* Print the entry */
    PrintEvent(pmDecodedDataFile,
	       lEventCPUID,
	       lEventID,
	       &lEventTime,
	       lCurrentPID,
	       lEventDataSize,
	       lEventString,
	       &lProperties,
	       pmOptions);
    }
}
